import dataclasses
from collections import defaultdict
from datetime import datetime

from recolul.duration import Duration
from recolul.errors import NoClockInError
from recolul.recoru.attendance_chart import AttendanceChart, ChartRow, ChartRowEntry

_MIN_HOURS_FOR_MANDATORY_BREAK = Duration(6 * 60)


def until_today(attendance_chart: AttendanceChart) -> AttendanceChart:
    """Return a slice of the attendance chart that only contains rows until today"""
    current_day_of_month = datetime.now().day
    return [
        row for row in attendance_chart
        if row.day_of_month <= current_day_of_month
    ]


def get_entry_work_time(entry: ChartRowEntry) -> Duration:
    """
    Get work time from the column if available,
    else calculate it from clock-in time and current time
    """
    category = entry.category
    if category.startswith("Half Day Leave"):
        return Duration(4 * 60)
    if category.endswith(("Leave", "Leagve")):
        return Duration(8 * 60)

    if not (raw_clock_in_time := entry.clock_in_time):
        return Duration(0)
    clock_in_time = Duration.parse(raw_clock_in_time)

    if raw_clock_out_time := entry.clock_out_time:
        clock_out_time = Duration.parse(raw_clock_out_time)
    else:
        # Current day
        clock_out_time = Duration.now()

    if clock_out_time < clock_in_time:
        # After midnight
        clock_out_time += Duration(24 * 60)

    work_time = clock_out_time - clock_in_time
    break_time = (
        Duration(60) if work_time >= _MIN_HOURS_FOR_MANDATORY_BREAK
        else Duration(0)
    )
    return work_time - break_time


def get_row_work_time(row: ChartRow) -> Duration:
    total_work_time = Duration()
    for entry in row.entries:
        total_work_time += get_entry_work_time(entry)
    return total_work_time


def get_overtime_history(attendance_chart: AttendanceChart) -> tuple[list[str], list[Duration], dict[str, Duration]]:
    days = []
    overtime_history = []
    total_workplace_times = defaultdict(Duration)
    for row in attendance_chart:
        day = row.day.text
        if not _is_working_day(row):
            required_time = Duration(0)
        else:
            required_time = Duration(8 * 60)

        row_work_time = Duration()
        for entry in row.entries:
            entry_work_time = get_entry_work_time(entry)
            row_work_time += entry_work_time

            workplace = entry.workplace or "HF Bldg."  # Workplace is empty for paid leaves
            total_workplace_times[workplace] += entry_work_time

        if not (required_time or row_work_time):
            # Can have work time during holidays
            continue

        days.append(day)
        overtime_history.append(row_work_time - required_time)

    return days, overtime_history, total_workplace_times


def get_overtime_balance(attendance_chart: AttendanceChart) -> tuple[Duration, dict[str, Duration]]:
    _, history, total_workplace_times = get_overtime_history(attendance_chart)
    return sum(history, Duration()), total_workplace_times


@dataclasses.dataclass
class LeaveTime:
    includes_break: bool
    min_time: Duration
    max_time: Duration | None = None


def get_leave_time(attendance_chart: AttendanceChart) -> list[LeaveTime]:
    day_base_hours = Duration(8 * 60)
    overtime_balance, _ = get_overtime_balance(attendance_chart[:-1])

    last_row = attendance_chart[-1]
    last_clock_in = None
    for entry in last_row.entries:
        if entry.clock_in_time and not entry.clock_out_time:
            # In progress
            last_clock_in = Duration.parse(entry.clock_in_time)
        else:
            # Complete entry
            overtime_balance += get_entry_work_time(entry)
    if not last_clock_in:
        raise NoClockInError()

    required_today = day_base_hours - overtime_balance
    leave_time_without_break = last_clock_in + required_today
    leave_time_with_break = last_clock_in + required_today + Duration(60)
    if required_today > _MIN_HOURS_FOR_MANDATORY_BREAK:
        # When more than 6 hours must be achieved during the day,
        # add a mandatory 1-hour break time.
        return [
            LeaveTime(includes_break=True, min_time=leave_time_with_break)
        ]
    if required_today > _MIN_HOURS_FOR_MANDATORY_BREAK - Duration(60):
        # When the required time is between 5 and 6 hours, there is a first interval where
        # the overtime balance becomes positive, and then it becomes negative again when 6
        # hours have been worked because an hour is subtracted for break time.
        first_leave_time = LeaveTime(
            includes_break=False,
            min_time=leave_time_without_break,
            max_time=last_clock_in + _MIN_HOURS_FOR_MANDATORY_BREAK
        )
        second_leave_time = LeaveTime(includes_break=True, min_time=leave_time_with_break)
        return [first_leave_time, second_leave_time]
    return [
            LeaveTime(includes_break=False, min_time=leave_time_without_break)
        ]


def count_working_days(attendance_chart: AttendanceChart) -> int:
    return sum(
        1
        for row in attendance_chart
        if _is_working_day(row)
    )


def _is_working_day(row: ChartRow) -> bool:
    return row.day.text and row.day.color not in ["blue", "red"] or "swap day" in row.memo.lower()



import plotly.graph_objects as go

from recolul.duration import Duration


def plot_overtime_balance_history(days: list[str], overtime_history: list[Duration]) -> None:
    cumulative_overtime_history = []
    for overtime in overtime_history:
        last_cumulative_overtime = (
            cumulative_overtime_history[-1] if cumulative_overtime_history
            else Duration()
        )
        cumulative_overtime_history.append(last_cumulative_overtime + overtime)

    fig = go.Figure(
            data=go.Scatter(
                x=days,
                y=[duration.minutes for duration in cumulative_overtime_history],
                text=[str(duration) for duration in cumulative_overtime_history],
                hovertemplate="%{x} %{text}<extra></extra>",
                mode="lines+markers"
            )
    )
    fig.update_layout(
        yaxis_title="Overtime balance (minutes)"
    )
    fig.show()



import argparse
import sys
from getpass import getpass

from recolul import __version__, plotting, time
from recolul.config import Config
from recolul.duration import Duration
from recolul.errors import NoClockInError
from recolul.recoru.attendance_chart import AttendanceChart
from recolul.recoru.recoru_session import RecoruSession
from recolul.time import get_entry_work_time, get_row_work_time, until_today


def balance(exclude_last_day: bool) -> None:
    full_attendance_chart = _get_attendance_chart()
    attendance_chart = until_today(full_attendance_chart)
    if exclude_last_day and len(attendance_chart) > 1:
        attendance_chart = attendance_chart[:-1]
    overtime_balance, total_workplace_times = time.get_overtime_balance(attendance_chart)
    print(f"Monthly overtime balance: {overtime_balance}")
    print(f"Total time per workplace:")
    for workplace, total_work_time in total_workplace_times.items():
        print(f"  {workplace}: {total_work_time}")
    print(f"Maximum WFH time this month: {Duration(60) * time.count_working_days(full_attendance_chart)}")

    if exclude_last_day:
        return

    last_day = attendance_chart[-1]
    print(f"\nLast day {last_day.day.text}")
    print(f"  Clock-in: {max(entry.clock_in_time for entry in last_day.entries)}")
    print(f"  Working hours: {get_row_work_time(last_day)}")


def when_to_leave() -> None:
    attendance_chart = until_today(_get_attendance_chart())
    try:
        leave_times = time.get_leave_time(attendance_chart)
    except NoClockInError:
        print("You have already clocked out.")
        return

    if len(leave_times) == 1:
        break_msg = "(includes a 1-hour break)" if leave_times[0].includes_break else "(break time not included)"
        print(f"Leave at {leave_times[0].min_time} to avoid overtime {break_msg}.")
    else:
        print(
            f"Leave between {leave_times[0].min_time} and {leave_times[0].max_time}, or """
            f"after {leave_times[1].min_time}."
        )


def update_config() -> None:
    """Needs to match Config.load"""
    recoru_contract_id = input("recoru.contractId: ")
    recoru_auth_id = input("recoru.authId: ")
    recoru_password = getpass("recoru.password: ")
    config = Config(
        recoru_contract_id=recoru_contract_id,
        recoru_auth_id=recoru_auth_id,
        recoru_password=recoru_password
    )
    config.save()


def graph(exclude_last_day: bool) -> None:
    attendance_chart = until_today(_get_attendance_chart())
    if exclude_last_day and len(attendance_chart) > 1:
        attendance_chart = attendance_chart[:-1]
    days, history, _ = time.get_overtime_history(attendance_chart)
    plotting.plot_overtime_balance_history(days, history)


def main() -> None:
    parser = argparse.ArgumentParser(prog="recolul")
    parser.add_argument("-v", "--version", action="version", version=__version__)
    subparsers = parser.add_subparsers(dest="command", required=True)

    balance_parser = subparsers.add_parser("balance", help="Calculate overtime balance")
    balance_parser.add_argument(
        "--exclude-last-day",
        action="store_true",
        help="Exclude last/current day from the calculation"
    )

    subparsers.add_parser("when", help="Calculate at which time to leave to avoid overtime this month")

    subparsers.add_parser("config", help="Init or update config")

    graph_parser = subparsers.add_parser("graph", help="Display a graph of overtime balance over the month")
    graph_parser.add_argument(
        "--exclude-last-day",
        action="store_true",
        help="Exclude last/current day from the graph"
    )

    args = parser.parse_args(sys.argv[1:])
    match args.command:
        case "balance":
            balance(exclude_last_day=args.exclude_last_day)
        case "when":
            when_to_leave()
        case "config":
            update_config()
        case "graph":
            graph(exclude_last_day=args.exclude_last_day)


def _get_attendance_chart() -> AttendanceChart:
    config = Config.from_env() or Config.load()
    if not config:
        raise RuntimeError(f"No config found")

    with RecoruSession(
        contract_id=config.recoru_contract_id,
        auth_id=config.recoru_auth_id,
        password=config.recoru_password
    ) as recoru_session:
        attendance_chart = recoru_session.get_attendance_chart()

    return attendance_chart


if __name__ == "__main__":
    main()



import re
from enum import Enum
from typing import TypeAlias

from bs4 import Tag


class ChartColumn(str, Enum):
    DATE = "日付"
    WORKPLACE = "作業場所"
    CATEGORY = "勤務区分"
    START = "開始"
    END = "終了"
    WORK_TIME = "労働時間"
    MEMO = "メモ"


class ChartHeader:
    """Header of the attendance chart"""
    def __init__(self, tag: Tag):
        self._column_indices: dict[ChartColumn, int] = {
            ChartCell(column).text.strip(): i
            for i, column in enumerate(tag.find_all("td", recursive=False))
        }

    def get_column_index(self, column: ChartColumn) -> int:
        return self._column_indices[column]

    def has_column(self, column: ChartColumn) -> bool:
        return column in self._column_indices


class ChartCell:
    """Single cell of the attendance chart"""
    def __init__(self, tag: Tag):
        self._tag = tag

    @property
    def text(self) -> str:
        return self._tag.text.strip()

    @property
    def color(self) -> str:
        label = self._tag.find("label", recursive=False)
        if not label:
            return ""
        return label.attrs\
            .get("style", "")\
            .removeprefix("color: ")\
            .removesuffix(";")


class ChartRowEntry:
    """Sub-row of the attendance chart"""
    def __init__(self, header: ChartHeader, tag: Tag):
        self._header = header
        self._tag = tag

    def __getitem__(self, column: ChartColumn) -> ChartCell:
        column_index = self._header.get_column_index(column)
        column = self._tag.select_one(f"td:nth-child({column_index + 1})", recursive=False)
        return ChartCell(column)

    @property
    def day(self) -> ChartCell:
        return self[ChartColumn.DATE]

    @property
    def workplace(self) -> str:
        return self[ChartColumn.WORKPLACE].text

    @property
    def category(self) -> str:
        return self[ChartColumn.CATEGORY].text

    @property
    def clock_in_time(self) -> str:
        return self[ChartColumn.START].text

    @property
    def clock_out_time(self) -> str:
        return self[ChartColumn.END].text

    @property
    def work_time(self) -> str:
        return self[ChartColumn.WORK_TIME].text

    @property
    def memo(self) -> str:
        return self[ChartColumn.MEMO].text


class ChartRow:
    """Row of the attendance chart"""
    _date_regex = re.compile(r"^(\d{1,2})\/(\d{1,2})\(.\)$")

    def __init__(self, entries: list[ChartRowEntry]):
        assert entries, "Empty ChartRow"
        self._entries = entries

    @property
    def entries(self) -> list[ChartRowEntry]:
        return self._entries

    @property
    def day(self) -> ChartCell:
        return self._entries[0][ChartColumn.DATE]

    @property
    def day_of_month(self) -> int:
        match = ChartRow._date_regex.match(self._entries[0].day.text)
        if not match:
            return 0
        return int(match.group(2))

    @property
    def memo(self) -> str:
        return self._entries[0][ChartColumn.MEMO].text


AttendanceChart: TypeAlias = list[ChartRow]



import requests
from bs4 import BeautifulSoup

from recolul.recoru.attendance_chart import AttendanceChart, ChartHeader, ChartRow, ChartRowEntry


class RecoruSession:
    def __init__(self, contract_id: str, auth_id: str, password: str):
        self._contract_id: str = contract_id
        self._auth_id: str = auth_id
        self._password: str = password

        self._session: requests.Session | None = None

    def __enter__(self):
        self._session = requests.Session()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._session.close()
        self._session = None

    @property
    def session(self) -> requests.Session:
        assert self._session, "RecoruSession should be used as a context manager"
        return self._session

    def get_attendance_chart(self) -> AttendanceChart:
        self._login()

        response = self.session.post("https://app.recoru.in/ap/home/loadAttendanceChartGadget")
        response.raise_for_status()
        return self._parse_attendance_chart(response.text)

    @classmethod
    def read_attendance_chart_file(cls, path: str) -> AttendanceChart:
        """Used for testing"""

        with open(path, "rt") as attendance_chart_file:
            text = attendance_chart_file.read()
        return cls._parse_attendance_chart(text)

    def _login(self):
        url = "https://app.recoru.in/ap/login"
        form_data = {
            "contractId": self._contract_id,
            "authId": self._auth_id,
            "password": self._password
        }
        response = self.session.post(url, data=form_data)
        response.raise_for_status()

    @staticmethod
    def _parse_attendance_chart(text: str) -> AttendanceChart:
        soup = BeautifulSoup(text, "html.parser")
        table = soup.select_one("#ID-attendanceChartGadgetTable")

        table_header = table.select_one("thead > tr", recursive=False)
        header = ChartHeader(table_header)

        chart_rows = []
        current_row_entries = []
        table_body = table.find("tbody", recursive=False)
        for row in table_body.find_all("tr", recursive=False):
            entry = ChartRowEntry(header, row)
            if entry.day.text:  # New row
                # Append previous row
                if current_row_entries:
                    chart_rows.append(ChartRow(current_row_entries))
                current_row_entries = [entry]
            else:  # Row with multiple entries
                current_row_entries.append(entry)
        if current_row_entries:
            chart_rows.append(ChartRow(current_row_entries))
        return chart_rows



import configparser
import os.path
from dataclasses import dataclass

CONFIG_PATH = os.path.realpath(f"{__file__}/../config.ini")


@dataclass
class Config:
    recoru_contract_id: str
    recoru_auth_id: str
    recoru_password: str

    @classmethod
    def from_env(cls):
        try:
            return cls(
                recoru_contract_id=os.environ["RECORU_CONTRACT_ID"],
                recoru_auth_id=os.getenv("RECORU_AUTH_ID"),
                recoru_password=os.environ["RECORU_PASSWORD"]
            )
        except KeyError:
            return None

    @classmethod
    def load(cls):
        if not os.path.isfile(CONFIG_PATH):
            return None

        config = configparser.ConfigParser(interpolation=None)
        config.read(CONFIG_PATH)
        return cls(
            recoru_contract_id=config["recoru"]["contractId"],
            recoru_auth_id=config["recoru"]["authId"],
            recoru_password=config["recoru"]["password"]
        )

    def save(self):
        config = configparser.ConfigParser(interpolation=None)
        config["recoru"] = {
            "authId": self.recoru_auth_id,
            "contractId": self.recoru_contract_id,
            "password": self.recoru_password
        }
        with open(CONFIG_PATH, "w") as config_file:
            config.write(config_file)



from datetime import datetime


class Duration:
    @classmethod
    def parse(cls, duration: str):
        if duration == "":
            return cls(0)
        hours, minutes = duration.split(":")
        total_minutes = 60 * int(hours) + int(minutes)
        return cls(total_minutes)

    @classmethod
    def now(cls):
        return cls.parse(datetime.now().strftime("%H:%M"))

    def __init__(self, minutes: int = 0):
        self.minutes: int = minutes

    def __repr__(self) -> str:
        return f"{'-' if self.minutes < 0 else ''}{abs(self.minutes) // 60:02}:{abs(self.minutes) % 60:02}"

    __str__ = __repr__

    def __eq__(self, other):
        return self.minutes == other.minutes

    def __neq__(self, other):
        return self.minutes != other.minutes

    def __add__(self, other):
        return Duration(self.minutes + other.minutes)

    def __sub__(self, other):
        return Duration(self.minutes - other.minutes)

    def __neg__(self):
        return Duration(-self.minutes)

    def __mul__(self, other):
        return Duration(other * self.minutes)

    def __lt__(self, other):
        return self.minutes < other.minutes

    def __gt__(self, other):
        return self.minutes > other.minutes

    def __le__(self, other):
        return self.minutes <= other.minutes

    def __ge__(self, other):
        return self.minutes >= other.minutes

    def __abs__(self):
        return Duration(abs(self.minutes))

    def __bool__(self):
        return self.minutes > 0
